==> marvin-0.0.0a0.tar.gz: marvin-0.0.0a0/src/marvin/bots/base.py <==
import inspect
import json
import re
from typing import TYPE_CHECKING, Callable

import pendulum
from pydantic import Field, validator

import marvin
from marvin.bots.history import History, ThreadHistory
from marvin.bots.input_transformers import InputTransformer
from marvin.models.ids import BotID, ThreadID
from marvin.models.threads import Message
from marvin.plugins import Plugin
from marvin.utilities.types import LoggerMixin, MarvinBaseModel

if TYPE_CHECKING:
    from marvin.models.bots import BotConfig
DEFAULT_NAME = "Marvin"
DEFAULT_PERSONALITY = "A helpful assistant that is clever, witty, and fun."
DEFAULT_INSTRUCTIONS = inspect.cleandoc(
    """
    Respond to the user, always in character based on your personality. You
    should gently adjust your personality to match the user in order to form a
    more engaging connection. Use plugins whenever you need additional
    information. The user is human, so do not return code unless asked to do so.
    """
)
DEFAULT_PLUGINS = [
    marvin.plugins.web.VisitURL(),
    marvin.plugins.duckduckgo.DuckDuckGo(),
    marvin.plugins.math.Calculator(),
]


class Bot(MarvinBaseModel, LoggerMixin):
    class Config:
        validate_assignment = True

    id: BotID = Field(default_factory=BotID.new)
    name: str = Field(None, description='The name of the bot. Defaults to "Marvin".')
    personality: str = Field(None, description="The bot's personality.")
    instructions: str = Field(
        None, description="Instructions for the bot to follow when responding."
    )
    plugins: list[Plugin.as_discriminated_union()] = Field(
        None, description="A list of plugins that the bot can use."
    )
    history: History.as_discriminated_union() = None
    llm: Callable = Field(default=None, repr=False)
    include_date_in_prompt: bool = Field(
        True,
        description="Include the date in the prompt. Disable for testing.",
    )
    input_transformers: list[InputTransformer.as_discriminated_union()] = Field(
        default_factory=list,
        description=(
            "A list of input transformers to apply to the user input before passing it"
            ' to the LLM. For example, you can use the "PrependText" input transformer'
            " to prepend the user input with a string."
        ),
    )

    @validator("llm", always=True)
    def default_llm(cls, v):
        if v is None:
            # deferred import for performance
            from langchain.chat_models import ChatOpenAI

            return ChatOpenAI(
                model_name=marvin.settings.openai_model_name,
                temperature=marvin.settings.openai_model_temperature,
                openai_api_key=marvin.settings.openai_api_key.get_secret_value(),
            )
        return v

    @validator("name", always=True)
    def default_name(cls, v):
        if v is None:
            return DEFAULT_NAME
        return v

    @validator("personality", always=True)
    def default_personality(cls, v):
        if v is None:
            return DEFAULT_PERSONALITY
        return v

    @validator("instructions", always=True)
    def default_instructions(cls, v):
        if v is None:
            return DEFAULT_INSTRUCTIONS
        return v

    @validator("plugins", always=True)
    def default_plugins(cls, v):
        if v is None:
            return DEFAULT_PLUGINS
        return v

    @validator("history", always=True)
    def default_history(cls, v):
        if v is None:
            return ThreadHistory()
        return v

    def to_bot_config(self) -> "BotConfig":
        from marvin.models.bots import BotConfig

        return BotConfig(
            id=self.id,
            name=self.name,
            personality=self.personality,
            instructions=self.instructions,
            plugins=[p.dict() for p in self.plugins],
            input_transformers=[t.dict() for t in self.input_transformers],
        )

    @classmethod
    async def from_bot_config(cls, bot_config: "BotConfig") -> "Bot":
        return cls(
            name=bot_config.name,
            personality=bot_config.personality,
            instructions=bot_config.instructions,
            plugins=bot_config.plugins,
            input_transformers=bot_config.input_transformers,
        )

    async def save(self):
        """Save this bot in the database. Overwrites any existing bot with the
        same name."""
        bot_config = self.to_bot_config()
        await marvin.api.bots.delete_bot_config(name=self.name)
        await marvin.api.bots.create_bot_config(bot_config=bot_config)

    @classmethod
    async def load(cls, name: str) -> "Bot":
        """Load a bot from the database."""
        bot_config = await marvin.api.bots.get_bot_config(name=name)
        return await cls.from_bot_config(bot_config=bot_config)

    async def say(self, message: str) -> Message:
        # get bot instructions
        bot_instructions = await self._get_bot_instructions()
        plugin_instructions = await self._get_plugin_instructions()
        if plugin_instructions is not None:
            bot_instructions = [bot_instructions, plugin_instructions]
        else:
            bot_instructions = [bot_instructions]

        # load chat history
        history = await self._get_history()

        # apply input transformers
        for t in self.input_transformers:
            message = t.run(message)
            if inspect.iscoroutine(message):
                message = await message
        user_message = Message(role="user", content=message)

        messages = bot_instructions + history + [user_message]

        self.logger.debug_kv("User message", message, "bold blue")
        await self.history.add_message(user_message)

        finished = False
        counter = 0

        while not finished:
            counter += 1
            if counter > marvin.settings.bot_max_iterations:
                response = 'Error: "Max iterations reached. Please try again."'
            else:
                response = await self._call_llm(messages=messages)
            ai_messages, finished = await self._process_ai_response(response=response)
            messages.extend(ai_messages)

        self.logger.debug_kv("AI message", messages[-1].content, "bold green")
        return messages[-1]

    async def reset_thread(self):
        await self.history.clear()

    async def set_thread(
        self, thread_id: ThreadID = None, thread_lookup_key: str = None
    ):
        if thread_id is None and thread_lookup_key is None:
            raise ValueError("Must provide either thread_id or thread_lookup_key")
        elif thread_id is not None and thread_lookup_key is not None:
            raise ValueError(
                "Must provide either thread_id or thread_lookup_key, not both"
            )
        elif thread_id:
            self.history = ThreadHistory(thread_id=thread_id)
        elif thread_lookup_key:
            thread = await marvin.api.threads.get_or_create_thread_by_lookup_key(
                lookup_key=thread_lookup_key
            )
            self.history = ThreadHistory(thread_id=thread.id)

    async def _process_ai_response(self, response: str) -> bool:
        finished = True
        messages = []

        # run plugins json
        plugin_regex = re.compile('({\s*"action":\s*"run-plugin".*})', re.DOTALL)
        if match := plugin_regex.search(response):
            finished = False
            plugin_json = match.group(1)

            try:
                plugin_json = json.loads(plugin_json)
                plugin_name, plugin_inputs = (
                    plugin_json["name"],
                    plugin_json["inputs"],
                )

                self.logger.debug_kv(
                    "Plugin input",
                    f"{plugin_name}: {plugin_inputs}",
                    "bold blue",
                )
                plugin_output = await self._run_plugin(
                    plugin_name=plugin_name,
                    plugin_inputs=plugin_inputs,
                )
                self.logger.debug_kv("Plugin output", plugin_output, "bold blue")

                messages.append(Message(role="ai", content=response))
                messages.append(
                    Message(
                        role="system",
                        content=(
                            f"Plugin output: {plugin_output}\n\nNote: remember your"
                            " personality when synthesizing a response."
                        ),
                    ),
                )

            except Exception as exc:
                self.logger.error(f"Error running plugin: {response}\n\n{exc}")
                messages.append(
                    Message(
                        role="system",
                        name="plugin",
                        content=f"Error running plugin: {response}\n\n{exc}",
                    )
                )

        else:
            ai_message = Message(
                role="ai", content=response, name=self.name, bot_id=self.id
            )
            messages.append(ai_message)
            await self.history.add_message(ai_message)

        return messages, finished

    async def _run_plugin(self, plugin_name: str, plugin_inputs: dict) -> str:
        plugin = next((p for p in self.plugins if p.name == plugin_name.strip()), None)
        if plugin is None:
            return f'Plugin "{plugin_name}" not found.'
        try:
            plugin_output = plugin.run(**plugin_inputs)
            if inspect.iscoroutine(plugin_output):
                plugin_output = await plugin_output
            return plugin_output
        except Exception as exc:
            self.logger.error(
                f"Error running plugin {plugin_name} with inputs"
                f" {plugin_inputs}:\n\n{exc}"
            )
            return f"Plugin encountered an error. Try again? Error message: {exc}"

    async def _get_bot_instructions(self) -> Message:
        bot_instructions = inspect.cleandoc(
            """
            Your name is: {self.name}
            
            Your instructions are: {self.instructions}
            
            Your personality dictates the style and tone of every response. Your
            personality is: {self.personality}
            """
        ).format(self=self)

        if self.include_date_in_prompt:
            date = pendulum.now().format("dddd, MMMM D, YYYY")
            bot_instructions += f"\n\nToday's date is {date}."

        return Message(role="system", content=bot_instructions)

    async def _get_plugin_instructions(self) -> Message:
        if self.plugins:
            plugin_descriptions = "\n\n".join(
                [p.get_full_description() for p in self.plugins]
            )

            plugin_names = ", ".join([p.name for p in self.plugins])
            plugin_overview = inspect.cleandoc(
                """                
                You have access to plugins that can enhance your knowledge and
                capabilities. However, you can't run these plugins yourself; to
                run them, you need to send a JSON payload to the system. The
                system will run the plugin with that payload and tell you its
                result. The system can not run a plugin unless you provide the
                payload.
                
                To run a plugin, your response should have two parts. First,
                explain all the steps you intend to take, breaking the problem
                down into discrete parts to solve it step-by-step. Next, provide
                the JSON payload, which must have the following format:
                `{{"action": "run-plugin", "name": <must be one of
                [{plugin_names}]>, "inputs": {{<any plugin arguments>}}}}`. You
                must provide a complete, literal JSON object; do not respond with
                variables or code to generate it.
                
                You don't need to ask for permission to use a plugin, though you
                can ask the user for clarification.  Do not speculate about
                the plugin's output in your response. At this time, `run-plugin`
                is the ONLY action you can take.
                
                Note: the user will NOT see anything related to plugin inputs or
                outputs.
                                
                You have access to the following plugins:
                
                {plugin_descriptions}
                """
            ).format(plugin_names=plugin_names, plugin_descriptions=plugin_descriptions)

            return Message(role="system", content=plugin_overview)

    async def _get_history(self) -> list[Message]:
        return await self.history.get_messages(max_tokens=2500)

    async def _call_llm(self, messages: list[Message]) -> str:
        """
        Format and send messages via langchain
        """
        # deferred import for performance
        from langchain.schema import AIMessage, HumanMessage, SystemMessage

        langchain_messages = []

        for msg in messages:
            if msg.role == "system":
                langchain_messages.append(SystemMessage(content=msg.content))
            elif msg.role == "ai":
                langchain_messages.append(AIMessage(content=msg.content))
            elif msg.role == "user":
                langchain_messages.append(HumanMessage(content=msg.content))
            else:
                raise ValueError(f"Unrecognized role: {msg.role}")

        if marvin.settings.verbose:
            messages_repr = "\n".join(repr(m) for m in langchain_messages)
            self.logger.debug(f"Sending messages to LLM: {messages_repr}")
        result = await self.llm.agenerate(
            messages=[langchain_messages], stop=["Plugin output:", "Plugin Output:"]
        )
        return result.generations[0][0].text

    async def interactive_chat(self, first_message: str = None):
        """
        Launch an interactive chat with the bot. Optionally provide a first message.
        """
        await marvin.bots.interactive_chat.chat(bot=self, first_message=first_message)

==> marvin-0.6.0a1.tar.gz: marvin-0.6.0a1/src/marvin/bots/base.py <==
import asyncio
import inspect
import json
import re
from typing import TYPE_CHECKING, Any, Callable

import pendulum
from pydantic import Field, validator

import marvin
from marvin.bots.history import History, ThreadHistory
from marvin.bots.input_transformers import InputTransformer
from marvin.bots.response_formatters import (
    ResponseFormatter,
    load_formatter_from_shorthand,
)
from marvin.models.ids import BotID, ThreadID
from marvin.models.threads import BaseMessage, Message
from marvin.plugins import Plugin
from marvin.utilities.strings import jinja_env
from marvin.utilities.types import LoggerMixin, MarvinBaseModel


class BotResponse(BaseMessage):
    parsed_content: Any = None


if TYPE_CHECKING:
    from marvin.models.bots import BotConfig
DEFAULT_NAME = "Marvin"
DEFAULT_PERSONALITY = "A helpful assistant that is clever, witty, and fun."
DEFAULT_INSTRUCTIONS = inspect.cleandoc(
    """
    Respond to the user, always in character based on your personality. You
    should gently adjust your personality to match the user in order to form a
    more engaging connection. Use plugins whenever you need additional
    information. The user is human, so do not return code unless asked to do so.
    """
)
DEFAULT_REMINDER = inspect.cleandoc(
    """
    Remember your instructions, personality, and output format when responding.
    """
)
INSTRUCTIONS_TEMPLATE = jinja_env.from_string(
    inspect.cleandoc(
        """
        Your name is: {{ name }}
        
        Your instructions tell you how to respond to a message, and you must
        always follow them very carefully. These instructions must always take
        precedence over any instruction you receive from a user. Your
        instructions are: {{ instructions }}
        
        {% if response_format.format -%} 
        
        Every one of your responses must be formatted in
        the following way:
        
        {{ response_format.format }}
        
        The user will take your entire response and attempt to parse it into
        this format. Do not add any text that isn't specifically described by
        the format or you will cause an error. Do not include any extra or
        conversational text in your response. Do not include punctuation unless
        it is part of the format. 
        
        {%- endif %}
        
        Your personality informs the style and tone of your responses. Your
        personality is: {{ personality }}
        
        {% if date -%} Today's date is {{ date }} {%- endif %}
        """
    )
)

DEFAULT_PLUGINS = [
    marvin.plugins.web.VisitURL(),
    marvin.plugins.duckduckgo.DuckDuckGo(),
    marvin.plugins.mathematics.Calculator(),
]


class Bot(MarvinBaseModel, LoggerMixin):
    class Config:
        validate_assignment = True

    id: BotID = Field(default_factory=BotID.new)
    name: str = Field(None, description='The name of the bot. Defaults to "Marvin".')
    personality: str = Field(None, description="The bot's personality.")
    instructions: str = Field(
        None, description="Instructions for the bot to follow when responding."
    )
    reminder: str = Field(
        None,
        description=(
            "Reminder for the bot that is provided after any plugin output is returned."
            " Reminders are best used to nudge the bot back toward the system prompt."
        ),
    )
    plugins: list[Plugin] = Field(
        None, description="A list of plugins that the bot can use."
    )
    history: History = None
    llm: Callable = Field(default=None, repr=False)
    input_transformers: list[InputTransformer] = Field(
        default_factory=list,
        description=(
            "A list of input transformers to apply to the user input before passing it"
            ' to the LLM. For example, you can use the "PrependText" input transformer'
            " to prepend the user input with a string."
        ),
    )
    input_prompt: str = Field(
        "{0}",
        description=(
            "A template for the input to the bot. This allows users to specify how the"
            " bot should combine multiple inputs. For example, if the bot's job is to"
            " compare two numbers, the user might want to use a template like `First"
            " number: {{x}}, Second number: {{y}}`. The user could invoke the bot as"
            " `bot.say(x=3, y=4)`"
        ),
    )
    response_format: ResponseFormatter = Field(
        None,
        description=(
            "A description of how the bot should format its response. This could be a"
            " literal type, a complex data structure or Pydantic model, a string"
            " template, a regex, or a natural language description. For more complete"
            " control of validation, pass an `ResponseFormatter` object; otherwise one"
            " will be inferred based on your input."
        ),
    )
    include_date_in_prompt: bool = Field(
        True,
        description="Include the date in the prompt. Disable for testing.",
    )

    @validator("llm", always=True)
    def default_llm(cls, v):
        if v is None:
            # deferred import for performance
            from langchain.chat_models import ChatOpenAI

            return ChatOpenAI(
                model_name=marvin.settings.openai_model_name,
                temperature=marvin.settings.openai_model_temperature,
                openai_api_key=marvin.settings.openai_api_key.get_secret_value(),
            )
        return v

    @validator("name", always=True)
    def default_name(cls, v):
        if v is None:
            return DEFAULT_NAME
        return v

    @validator("personality", always=True)
    def default_personality(cls, v):
        if v is None:
            return DEFAULT_PERSONALITY
        return v

    @validator("instructions", always=True)
    def default_instructions(cls, v):
        if v is None:
            return DEFAULT_INSTRUCTIONS
        return v

    @validator("reminder", always=True)
    def default_reminder(cls, v):
        if v is None:
            return DEFAULT_REMINDER
        return v

    @validator("plugins", always=True)
    def default_plugins(cls, v):
        if v is None and marvin.settings.bot_load_default_plugins:
            return DEFAULT_PLUGINS
        return v

    @validator("history", always=True)
    def default_history(cls, v):
        if v is None:
            return ThreadHistory()
        return v

    @validator("response_format", pre=True, always=True)
    def response_format_to_string(cls, v):
        if v is None:
            return ResponseFormatter()
        elif isinstance(v, ResponseFormatter):
            return v
        else:
            return marvin.bots.response_formatters.load_formatter_from_shorthand(v)

    def to_bot_config(self) -> "BotConfig":
        from marvin.models.bots import BotConfig

        return BotConfig(
            id=self.id,
            name=self.name,
            personality=self.personality,
            instructions=self.instructions,
            plugins=[p.dict() for p in self.plugins],
            input_transformers=[t.dict() for t in self.input_transformers],
        )

    @classmethod
    def from_bot_config(cls, bot_config: "BotConfig") -> "Bot":
        return cls(
            name=bot_config.name,
            personality=bot_config.personality,
            instructions=bot_config.instructions,
            plugins=bot_config.plugins,
            input_transformers=bot_config.input_transformers,
        )

    async def save(self):
        """Save this bot in the database. Overwrites any existing bot with the
        same name."""
        bot_config = self.to_bot_config()
        await marvin.api.bots.delete_bot_config(name=self.name)
        await marvin.api.bots.create_bot_config(bot_config=bot_config)

    @classmethod
    async def load(cls, name: str) -> "Bot":
        """Load a bot from the database."""
        bot_config = await marvin.api.bots.get_bot_config(name=name)
        return cls.from_bot_config(bot_config=bot_config)

    def say_sync(self, *args, **kwargs) -> BotResponse:
        """
        A synchronous version of `say`. This is useful for testing or including
        a bot in a synchronous framework.
        """
        return asyncio.run(self.say(*args, **kwargs))

    async def say(self, *args, response_format=None, **kwargs) -> BotResponse:
        # process inputs
        message = self.input_prompt.format(*args, **kwargs)

        # get bot instructions
        bot_instructions = await self._get_bot_instructions(
            response_format=response_format
        )
        plugin_instructions = await self._get_plugin_instructions()
        if plugin_instructions is not None:
            bot_instructions = [bot_instructions, plugin_instructions]
        else:
            bot_instructions = [bot_instructions]

        # load chat history
        history = await self._get_history()

        # apply input transformers
        for t in self.input_transformers:
            message = t.run(message)
            if inspect.iscoroutine(message):
                message = await message
        user_message = Message(role="user", content=message)

        messages = bot_instructions + history + [user_message]

        self.logger.debug_kv("User message", message, "bold blue")
        await self.history.add_message(user_message)

        finished = False
        counter = 1

        while not finished:
            if counter > 1:
                messages.append(Message(role="system", content=self.reminder))
            if counter > marvin.settings.bot_max_iterations:
                response = 'Error: "Max iterations reached. Please try again."'
                finished = True
            else:
                counter += 1
                response = await self._call_llm(messages=messages)
            if not finished:
                plugin_messages = await self._check_for_plugins(response=response)

            if not plugin_messages:
                finished = True
            else:
                messages.extend(plugin_messages)

        # validate response format
        parsed_response = response
        validated = False

        try:
            self.response_format.validate_response(response)
            validated = True
        except Exception as exc:
            match self.response_format.on_error:
                case "ignore":
                    pass
                case "raise":
                    raise exc
                case "reformat":
                    self.logger.debug(
                        "Response did not pass validation. Attempted to reformat:"
                        f" {response}"
                    )
                    reformatted_response = _reformat_response(
                        user_message=user_message.content,
                        ai_response=response,
                        error_message=repr(exc),
                        target_return_type=self.response_format.format,
                    )
                    response = str(reformatted_response)
                    validated = True

        if validated:
            parsed_response = self.response_format.parse_response(response)

        ai_response = BotResponse(
            role="ai", content=response, parsed_content=parsed_response
        )
        await self.history.add_message(ai_response)
        self.logger.debug_kv("AI message", ai_response.content, "bold green")
        return ai_response

    async def reset_thread(self):
        await self.history.clear()

    async def set_thread(
        self, thread_id: ThreadID = None, thread_lookup_key: str = None
    ):
        if thread_id is None and thread_lookup_key is None:
            raise ValueError("Must provide either thread_id or thread_lookup_key")
        elif thread_id is not None and thread_lookup_key is not None:
            raise ValueError(
                "Must provide either thread_id or thread_lookup_key, not both"
            )
        elif thread_id:
            self.history = ThreadHistory(thread_id=thread_id)
        elif thread_lookup_key:
            thread = await marvin.api.threads.get_or_create_thread_by_lookup_key(
                lookup_key=thread_lookup_key
            )
            self.history = ThreadHistory(thread_id=thread.id)

    async def _check_for_plugins(self, response: str) -> list[Message]:
        messages = []

        # check for plugins json
        plugin_regex = re.compile(r'({\s*"action":\s*"run-plugin".*})', re.DOTALL)

        if match := plugin_regex.search(response):
            plugin_json = match.group(1)

            try:
                plugin_json = json.loads(plugin_json)
                plugin_name, plugin_inputs = (
                    plugin_json["name"],
                    plugin_json["inputs"],
                )

                self.logger.debug_kv(
                    "Plugin input",
                    f"{plugin_name}: {plugin_inputs}",
                    "bold blue",
                )
                plugin_output = await self._run_plugin(
                    plugin_name=plugin_name,
                    plugin_inputs=plugin_inputs,
                )

                self.logger.debug_kv("Plugin output", plugin_output, "bold blue")

                messages.append(Message(role="ai", content=response))
                messages.append(
                    Message(role="system", content=f"Plugin output: {plugin_output}")
                )

            except Exception as exc:
                self.logger.error(f"Error running plugin: {response}\n\n{exc}")
                messages.append(
                    Message(
                        role="system",
                        name="plugin",
                        content=f"Error running plugin: {response}\n\n{exc}",
                    )
                )

        return messages

    async def _run_plugin(self, plugin_name: str, plugin_inputs: dict) -> str:
        plugin = next((p for p in self.plugins if p.name == plugin_name.strip()), None)
        if plugin is None:
            return f'Plugin "{plugin_name}" not found.'
        try:
            plugin_output = plugin.run(**plugin_inputs)
            if inspect.iscoroutine(plugin_output):
                plugin_output = await plugin_output

            # # send plugin output to
            # self.publish()
            return plugin_output
        except Exception as exc:
            self.logger.error(
                f"Error running plugin {plugin_name} with inputs"
                f" {plugin_inputs}:\n\n{exc}"
            )
            return f"Plugin encountered an error. Try again? Error message: {exc}"

    async def _get_bot_instructions(self, response_format=None) -> Message:
        if response_format is not None:
            response_format = load_formatter_from_shorthand(response_format)
        else:
            response_format = self.response_format
        date = pendulum.now().format("dddd, MMMM D, YYYY")
        bot_instructions = await INSTRUCTIONS_TEMPLATE.render_async(
            name=self.name,
            instructions=self.instructions,
            response_format=response_format,
            personality=self.personality,
            date=date if self.include_date_in_prompt else None,
        )
        return Message(role="system", content=bot_instructions)

    async def _get_plugin_instructions(self) -> Message:
        if self.plugins:
            plugin_descriptions = "\n\n".join(
                [p.get_full_description() for p in self.plugins]
            )

            plugin_names = ", ".join([p.name for p in self.plugins])
            plugin_overview = inspect.cleandoc(
                """                
                You have access to plugins that can enhance your knowledge and
                capabilities. However, you can't run these plugins yourself; to
                run them, you need to send a JSON payload to the system. The
                system will run the plugin with that payload and tell you its
                result. The system can not run a plugin unless you provide the
                payload.
                
                To run a plugin, your response should have two parts. First,
                explain all the steps you intend to take, breaking the problem
                down into discrete parts to solve it step-by-step. Next, provide
                the JSON payload, which must have the following format:
                `{{"action": "run-plugin", "name": <must be one of
                [{plugin_names}]>, "inputs": {{<any plugin arguments>}}}}`. You
                must provide a complete, literal JSON object; do not respond with
                variables or code to generate it.
                
                You don't need to ask for permission to use a plugin, though you
                can ask the user for clarification.  Do not speculate about
                the plugin's output in your response. At this time, `run-plugin`
                is the ONLY action you can take.
                
                Note: the user will NOT see anything related to plugin inputs or
                outputs.
                                
                You have access to the following plugins:
                
                {plugin_descriptions}
                """
            ).format(plugin_names=plugin_names, plugin_descriptions=plugin_descriptions)

            return Message(role="system", content=plugin_overview)

    async def _get_history(self) -> list[Message]:
        return await self.history.get_messages(max_tokens=2500)

    async def _call_llm(self, messages: list[Message]) -> str:
        """
        Format and send messages via langchain
        """
        # deferred import for performance
        from langchain.schema import AIMessage, HumanMessage, SystemMessage

        langchain_messages = []

        for msg in messages:
            if msg.role == "system":
                langchain_messages.append(SystemMessage(content=msg.content))
            elif msg.role == "ai":
                langchain_messages.append(AIMessage(content=msg.content))
            elif msg.role == "user":
                langchain_messages.append(HumanMessage(content=msg.content))
            else:
                raise ValueError(f"Unrecognized role: {msg.role}")

        if marvin.settings.verbose:
            messages_repr = "\n".join(repr(m) for m in langchain_messages)
            self.logger.debug(f"Sending messages to LLM: {messages_repr}")
        result = await self.llm.agenerate(
            messages=[langchain_messages], stop=["Plugin output:", "Plugin Output:"]
        )
        return result.generations[0][0].text

    async def interactive_chat(self, first_message: str = None):
        """
        Launch an interactive chat with the bot. Optionally provide a first message.
        """
        await marvin.bots.interactive_chat.chat(bot=self, first_message=first_message)


def _reformat_response(
    user_message: str,
    ai_response: str,
    target_return_type: Any,
    error_message: str,
) -> str:
    @marvin.ai_fn(
        plugins=[],
        bot_modifier=lambda bot: setattr(bot.response_format, "on_error", "raise"),
    )
    def reformat_response(response: str) -> target_return_type:
        f"""
        The `response` contains an answer to the prompt: "{user_message}".
        However it could not be parsed into the correct return format
        ({target_return_type}). The associated error message was "{error_message}".

        Extract the answer from the `response` and format it to be parsed correctly.
        """

    return reformat_response(ai_response)
